/*
 * plugin.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"
#include <Winamp/wa_ipc.h>

#define WA_TOGGLE_VIS 40192

#define plugin_is_mjpeg(p) \
  (CODEC_ID_MJPEG==(p)->play.ffi.vst->st->codec->codec_id)

#define plugin_is_video(p) \
  (0!=(p)->ps.general.vid&&NULL!=(p)->play.ffi.vst&&!plugin_is_mjpeg(p))

#define plugin_overflow(p) \
  (queue_overflow(&(p)->a.q)&&(!plugin_is_video(p)||queue_overflow(&(p)->v.q)))

plugin_t *plugin_init(plugin_t *p, In_Module *mod, pb_str_t *root,
    pb_str_t *home)
{
  p->mod=mod;
  p->root=root;
  p->home=home;

  prefs_load(&p->ps,home,mod->hMainWindow);

  p->state=STATE_RUN;

  p->hMutex=CreateMutex(
    NULL,   // _In_opt_  LPSECURITY_ATTRIBUTES lpMutexAttributes,
    FALSE,  // _In_      BOOL bInitialOwner,
    NULL    // _In_opt_  LPCTSTR lpName
  );

  if (NULL==p->hMutex)
    goto error;

  p->hEventWrite=CreateEvent(
    NULL,   // lpEventAttributes
    FALSE,  // bManualReset
    FALSE,  // bInitialState
    NULL    // lpName
  );

  if (NULL==p->hEventWrite)
    goto error;

  if (NULL==audio_init(&p->a,p))
    goto error;

  if (NULL==video_init(&p->v,p))
    goto error;

  return p;
error:
  plugin_cleanup(p);

  return NULL;
}

plugin_t *plugin_cleanup(plugin_t *p)
{
  video_cleanup(&p->v);
  audio_cleanup(&p->a);

  if (NULL!=p->hEventWrite)
    CloseHandle(p->hEventWrite);

  if (NULL!=p->hMutex)
    CloseHandle(p->hMutex);

  if (NULL==p->home)
    free(p->home);

  if (NULL==p->root)
    free(p->root);

  return p;
}

#if defined (PLUGIN_REALLOC) // {
plugin_t *plugin_start(plugin_t *p, const in_char *fn)
{
  if (NULL==plugin_realloc(p,fn))
    goto error;

  return p;
error:
  plugin_stop(p);

  return NULL;
}
#else // } {
plugin_t *plugin_start(plugin_t *p, const in_char *fn)
{
  HWND hMainWindow=p->mod->hMainWindow;
  int visrunning=SendMessage(hMainWindow,WM_WA_IPC,0,IPC_ISVISRUNNING);

  ZERO(&p->play);

  if (NULL==track_init(&p->play,fn,&p->ps,1))
    goto error;

  if (NULL==audio_realloc(&p->a,&p->play))
    goto error;

  p->is_video=plugin_is_video(p);

  if (p->is_video) {
    if (NULL==video_realloc(&p->v,&p->play))
      goto error;
  }
  else if (0!=p->ps.general.vis&&!visrunning) {
	if (p->v.s.open) {
      screen_close(&p->v.s);
      screen_open(&p->v.s,0,0,0.0);
      screen_close(&p->v.s);
    }

    PostMessage(hMainWindow,WM_COMMAND,WA_TOGGLE_VIS,0);
  }

  return p;
error:
  return NULL;
}
#endif // }

void plugin_stop(plugin_t *p, int killed)
{
  if (killed)
    plugin_killed(p);

  if (p->is_video)
    video_stop(&p->v);

  audio_stop(&p->a);
  track_cleanup(&p->play);
  ZERO(&p->play);
}

double plugin_get_time(plugin_t *p)
{
  Out_Module *outMod=p->mod->outMod;

  return p->a.s.time
      +outMod->GetOutputTime()
	  -outMod->GetWrittenTime(); 
}

void plugin_reset(plugin_t *p)
{
  p->state=STATE_RUN;
  audio_reset(&p->a);

  if (p->is_video)
    video_reset(&p->v);
}

plugin_t *plugin_run(plugin_t *p)
{
  plugin_reset(p);

  p->hThread=CreateThread(
    NULL,   // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
    0,      // _In_       SIZE_T dwStackSize,
    plugin_thread,
            // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
    p,      // _In_opt_   LPVOID lpParameter,
    0,      // _In_       DWORD dwCreationFlags,
    NULL    // _Out_opt_  LPDWORD lpThreadId
  );

  if (NULL==p->hThread)
    goto error;

  return p;
error:
  return NULL;
}

void plugin_pause(plugin_t *p)
{
  WaitForSingleObject(p->hMutex,INFINITE);

  if (STATE_RUN==p->state) {
    p->state=STATE_PAUSE;
	SetEvent(p->a.q.hEventRead);

    if (p->is_video)
	  SetEvent(p->v.q.hEventRead);

	SetEvent(p->hEventWrite);
  }
  
  ReleaseMutex(p->hMutex);
}

void plugin_unpause(plugin_t *p)
{
  WaitForSingleObject(p->hMutex,INFINITE);

  if (STATE_PAUSE==p->state) {
    p->state=STATE_RUN;
	SetEvent(p->a.q.hEventRead);

    if (p->is_video)
	  SetEvent(p->v.q.hEventRead);

	SetEvent(p->hEventWrite);
  }
  
  ReleaseMutex(p->hMutex);
}

void plugin_wait(plugin_t *p)
{
  if (NULL!=p->hThread) {
    WaitForSingleObject(p->hThread,INFINITE);
    CloseHandle(p->hThread);
    p->hThread=NULL;
  }
}

#if defined (PLUGIN_REALLOC) // {
plugin_t *plugin_realloc(plugin_t *p, const in_char *fn)
{
  HWND hMainWindow=p->mod->hMainWindow;
  int visrunning=SendMessage(hMainWindow,WM_WA_IPC,0,IPC_ISVISRUNNING);

  ZERO(&p->play);

  if (NULL==track_init(&p->play,fn,&p->ps,1))
    goto error;

  if (NULL==audio_realloc(&p->a,&p->play))
    goto error;

  p->is_video=plugin_is_video(p);

  if (p->is_video) {
    if (NULL==video_realloc(&p->v,&p->play))
      goto error;
  }
  else if (0!=p->ps.general.vis&&!visrunning) {
	if (p->v.s.open) {
      screen_close(&p->v.s);
      screen_open(&p->v.s,0,0,0.0);
      screen_close(&p->v.s);
    }

    PostMessage(hMainWindow,WM_COMMAND,WA_TOGGLE_VIS,0);
  }

  return p;
error:
  return NULL;
}
#endif // }

void plugin_kill(plugin_t *p)
{
  audio_kill(&p->a);

  if (p->is_video)
    video_kill(&p->v);

  WaitForSingleObject(p->hMutex,INFINITE);
  p->state=STATE_KILL;
  SetEvent(p->a.q.hEventRead);

  if (p->is_video)
    SetEvent(p->v.q.hEventRead);

  SetEvent(p->hEventWrite);
  ReleaseMutex(p->hMutex);
}

void plugin_eof(plugin_t *p)
{
  WaitForSingleObject(p->hMutex,INFINITE);

  if (STATE_RUN==p->state) {
    p->state=STATE_EOF;
    SetEvent(p->a.q.hEventRead);

    if (p->is_video)
      SetEvent(p->v.q.hEventRead);

    SetEvent(p->hEventWrite);
  }

  ReleaseMutex(p->hMutex);
}

void plugin_killed(plugin_t *p)
{
  HWND hMainWindow=p->mod->hMainWindow;
  int visrunning=SendMessage(hMainWindow,WM_WA_IPC,0,IPC_ISVISRUNNING);

  if (p->is_video&&!visrunning)
    screen_close(&p->v.s);
}

int plugin_seek(plugin_t *p, int64_t pos_ms)
{
  ffi_t *ffi=&p->play.ffi;
  // prefer then video stream for seeking.
  ffi_st_t *st=NULL==ffi->vst?ffi->ast:ffi->vst;
  int64_t seek_target=av_rescale_q(1000*pos_ms,AV_TIME_BASE_Q,
      st->st->time_base);
  int ret;

  plugin_kill(p);
  plugin_wait(p);

  avcodec_flush_buffers(ffi->ast->dc);

  if (p->is_video)
    avcodec_flush_buffers(ffi->vst->dc);

  ret=avformat_seek_file(ffi->ic,st->index,INT64_MIN,seek_target,
        INT64_MAX,0);

  if (NULL==plugin_run(p))
    goto error;

  return ret;
error:
  plugin_kill(p);
  plugin_stop(p,0);

  return -1;
}

void plugin_write(plugin_t *p, queue_t *q, AVPacket *pkt)
{
  WaitForSingleObject(p->hMutex,INFINITE);

  while (STATE_KILL!=p->state&&plugin_overflow(p)) {
    SignalObjectAndWait(p->hMutex,p->hEventWrite,INFINITE,FALSE);
    WaitForSingleObject(p->hMutex,INFINITE);
  }

  if (STATE_KILL!=p->state) {
    queue_write(q,pkt);
    SetEvent(q->hEventRead);
  }
  else
    av_free_packet(pkt);

  ReleaseMutex(p->hMutex);
}

DWORD WINAPI plugin_thread(LPVOID lpParam)
{
  plugin_t *p=lpParam;
  ffi_t *ffi=&p->play.ffi;
  ffi_st_t *ast=ffi->ast,*vst=ffi->vst;
  queue_t *aq=&p->a.q,*vq=&p->v.q;
  AVPacket pkt;

  // initialize packet, set data to NULL, let the demuxer fill it
  ZERO(&pkt);
  av_init_packet(&pkt);

  // start sub-threads
  if (NULL==audio_run(&p->a))
    goto cleanup_audio;

  if (p->is_video&&NULL==video_run(&p->v))
    goto cleanup_video;

  while (STATE_KILL!=p->state&&0<=av_read_frame(ffi->ic,&pkt)) {
    if (pkt.stream_index==ast->index)
      plugin_write(p,aq,&pkt);
    else if (p->is_video&&pkt.stream_index==vst->index)
      plugin_write(p,vq,&pkt);
    else
      av_free_packet(&pkt);
  }

  plugin_eof(p);

  // wait for sub-threads.
  if (p->is_video)
    video_wait(&p->v);
cleanup_video:
  audio_wait(&p->a);
cleanup_audio:
  return 0;
}
