/*
 * prefs.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include <Winamp/wa_ipc.h>
#include "prefdef.h"

///////////////////////////////////////////////////////////////////////////////
#define DEFAULT_PAGE 1

static prefs_t dps={
  // general
  .general.bps=16,
  .general.vis=1,
  .general.vid=1,

  // rg
  .rg.weight=0.5,
  .rg.preamp=0.0,
  .rg.comment=1,

  // audio
  .audio.queue_size=QUEUE_SIZE,

  // sox
  .sox.mode="auto",
  .sox.rate=44100,
  .sox.rate_min=1,
  .sox.rate_mul=1,
  .sox.rate_q="-v",
  .sox.dither="-s",

  // video
  .video.queue_size=QUEUE_SIZE,
  .video.min_drift=-800.0,
  .video.max_drift=8000.0,
};

///////////////////////////////////////////////////////////////////////////////
typedef struct page {
  prefs_general_t general;
  prefs_audio_t audio;
  prefs_video_t video;
} page_t;

static tcp_t tcp[]={
  { L"General",IDD_PREFS_GENERAL,GeneralDlgProc,offsetof(page_t,general) },
  { L"Audio",IDD_PREFS_AUDIO,AudioDlgProc,offsetof(page_t,audio) },
  { L"Video",IDD_PREFS_VIDEO,VideoDlgProc,offsetof(page_t,video) },
  { NULL,0,NULL }
};

enum { PREFS_NPAGES=(sizeof tcp)/(sizeof tcp[0])-1 };

typedef struct prefsdlg {
  prefs_t *ps;
  page_t page;
  tabs_t tc;
  HWND hPages[PREFS_NPAGES];
} priv_t;

///////////////////////////////////////////////////////////////////////////////
void prefs_read(prefs_t *ps, prefdef_t *pd)
{
  char *path=PS_UTF8(ps->path);
  char buf[128];

  switch (pd->type) {
  case PREF_INT:
    pref_int(pd,ps)=GetPrivateProfileIntA(pd->group,pd->name,
        pref_int(pd,&dps),path);
    break;
  case PREF_DOUBLE:
    sprintf(buf,"%f",pref_double(pd,&dps));
    GetPrivateProfileStringA(pd->group,pd->name,buf,buf,sizeof buf,path);
    pref_double(pd,ps)=atof(buf);
    break;
  case PREF_STRING:
    GetPrivateProfileStringA(pd->group,pd->name,pref_string(pd,&dps),
	    pref_string(pd,ps),pd->size,path);
    break;
  default:
    break;
  }
}

void prefs_write(prefs_t *ps, prefdef_t *pd)
{
  char *path=PS_UTF8(ps->path);
  char buf[128];

  WritePrivateProfileStringA(GENERAL,"exts",ps->exts.display.rep,path);

  switch (pd->type) {
  case PREF_INT:
    sprintf(buf,"%d",pref_int(pd,ps));
    WritePrivateProfileStringA(pd->group,pd->name,buf,path);
    break;
  case PREF_DOUBLE:
    sprintf(buf,"%f",pref_double(pd,ps));
    WritePrivateProfileStringA(pd->group,pd->name,buf,path);
    break;
  case PREF_STRING:
    WritePrivateProfileStringA(pd->group,pd->name,pref_string(pd,ps),path);
    break;
  default:
    break;
  }
}

///////////////////////////////////////////////////////////////////////////////
void prefs_load(prefs_t *ps, pb_str_t *home, HWND hMainWindow)
{
  char *dir=(char *)SendMessageW(hMainWindow,WM_WA_IPC,0,IPC_GETINIDIRECTORY);
  pb_str_t *p1=NULL;
  pb_str_t *p2=NULL;

  if (NULL==dir||0==*dir) {
    if (NULL==(ps->path=pb_str_cat_ps(home,_PS("\\in_ffsox.ini"),NULL)))
      goto cleanup;
  }
  else {
    if (NULL==(p1=pb_str_cre_cs(dir)))
      goto cleanup;

#if defined (PLUGIN_DIR) // {
    if (NULL==(p2=pb_str_cat_ps(p1,_PS("\\Plugins\\in_ffsox"),NULL)))
      goto cleanup;
#else // } {
    if (NULL==(p2=pb_str_cat_ps(p1,_PS("\\Plugins"),NULL)))
      goto cleanup;
#endif // }

    CreateDirectoryW(
      PS_TCS(p2),   // __in      LPCTSTR lpPathName,
      NULL          // __in_opt  LPSECURITY_ATTRIBUTES lpSecurityAttributes
    );

    if (NULL==(ps->path=pb_str_cat_ps(p2,_PS("\\in_ffsox.ini"),NULL)))
      goto cleanup;
  }

  prefs_general_load(ps);
  prefs_audio_load(ps);
  prefs_video_load(ps);
cleanup:
  if (NULL!=p2)
    free(p2);

  if (NULL!=p1)
    free(p1);
}

///////////////////////////////////////////////////////////////////////////////
static void PrefsDlgDispatch(priv_t *p, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  int i;

  for (i=0;i<PREFS_NPAGES;++i)
    SendMessage(p->hPages[i],uMsg,wParam,lParam);
}

BOOL CALLBACK PrefsDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  priv_t *p=WM_INITDIALOG==uMsg?(void *)lParam
      :(void *)GetWindowLongPtr(hDlg,GWLP_USERDATA);
  int n;
  char buf[32];

  switch(uMsg) {
  case WM_INITDIALOG:
    SetWindowLongPtr(hDlg,GWLP_USERDATA,lParam);
    SetWindowTextA(hDlg,CAPTION " Preferences");
    n=GetPrivateProfileIntA(GENERAL,"page",DEFAULT_PAGE,PS_UTF8(p->ps->path));
    tabs_init(&p->tc,hDlg,IDC_TAB,n-1,tcp,p->hPages,(char *)&p->page);
    return TRUE;
  case WM_NOTIFY:
    if (IDC_TAB==wParam&&TCN_SELCHANGE==((NMHDR *)lParam)->code) {
      tabs_change(&p->tc,hDlg);
      return TRUE;
    }

    break;
  case WM_COMMAND:
    switch(LOWORD(wParam)) {
    case IDOK:
      n=TabCtrl_GetCurSel(GetDlgItem(hDlg,p->tc.id)); 
	  sprintf(buf,"%d",1+n);
      WritePrivateProfileStringA(GENERAL,"page",buf,PS_UTF8(p->ps->path));
	  PrefsDlgDispatch(p,WM_COMMAND,IDOK,0);
      EndDialog(hDlg,IDOK);
      return TRUE;
    case IDCANCEL:
      EndDialog(hDlg,IDCANCEL);
      return TRUE;
    case IDC_BUTTON_DEFAULT:
	  PrefsDlgDispatch(p,WM_COMMAND,IDC_BUTTON_DEFAULT,(LPARAM)&dps);
      tabs_set_page(&p->tc,hDlg,DEFAULT_PAGE-1);
      return TRUE;
	default:
	  break;
    }

    break;
  default:
    break;
  }

  return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
void prefs_edit(prefs_t *ps, HWND hParent)
{
  priv_t p;

  ZERO(&p);
  p.ps=ps;
  p.page.general.ps=ps;
  p.page.audio.ps=ps;
  p.page.video.ps=ps;

  DialogBoxParamW(
    GetModuleHandleA(MODULE),       // __in_opt  HINSTANCE hInstance
    MAKEINTRESOURCEW(IDD_PREFS),    // __in LPCTSTR lpTemplateName
    hParent,                        // __in_opt  HWND hWndParent
    PrefsDlgProc,                   // __in_opt  DLGPROC lpDialogFunc
    (LPARAM)&p                      // __in LPARAM dwInitParam
  );
}
