/*
 * queue.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

queue_t *queue_init(queue_t *q, plugin_t *p, char *name, int size)
{
  q->p=p;
  q->hMutex=p->hMutex;
  q->hEventWrite=p->hEventWrite;
  q->name=name;

  q->hEventRead=CreateEvent(
    NULL,           // lpEventAttributes
    FALSE,          // bManualReset
    FALSE,          // bInitialState
    NULL            // lpName
  );

  if (NULL==q->hEventRead)
    goto error;

#if ! defined (QUEUE_FREE) // {
  q->free=&p->free;
#endif // }
  q->size=size;
  q->count=0;

  return q;
error:
  queue_cleanup(q);

  return NULL;
}

queue_t *queue_cleanup(queue_t *q)
{
  node_t *n;

  queue_flush(q);

#if defined (QUEUE_FREE) // {
  while (NULL!=(n=list_read(&q->free)))
    free(n);
#else // } {
  while (NULL!=(n=list_read(q->free)))
    free(n);
#endif // }

  if (NULL!=q->hEventRead)
    CloseHandle(q->hEventRead);

  return q;
}

void queue_reset(queue_t *q, int size)
{
  queue_flush(q);
  q->size=size;
}

void queue_flush(queue_t *q)
{
  node_t *n;

  while (0<q->count) {
    n=list_read(&q->active);
    av_free_packet(&n->pkt);
#if defined (QUEUE_FREE) // {
	list_write(&q->free,n);
#else // } {
	list_write(q->free,n);
#endif // }
    --q->count;
  }
}

// non-blocking.
int queue_write(queue_t *q, AVPacket *pkt)
{
  int state=-1;
  node_t *n;

#if defined (QUEUE_FREE) // {
  if (NULL==(n=list_read(&q->free))&&NULL==(n=malloc(sizeof *n)))
    goto error;
#else // } {
  if (NULL==(n=list_read(q->free))&&NULL==(n=malloc(sizeof *n)))
    goto error;
#endif // }

  n->pkt=*pkt;
  ZERO(pkt);
  av_init_packet(pkt);
  list_write(&q->active,n);
  ++q->count;
//fprintf(stderr,"%s written: %d.\n",q->name,q->count);
  state=0;
error:
  return state;
}

// blocking.
int queue_read(queue_t *q, AVPacket *pkt)
{
  int state=-1;
  node_t *n;
  plugin_t *p=q->p;

  WaitForSingleObject(p->hMutex,INFINITE);

  while (STATE_PAUSE==p->state||(STATE_RUN==p->state&&0==q->count)) {
    SignalObjectAndWait(p->hMutex,q->hEventRead,INFINITE,FALSE);
    WaitForSingleObject(p->hMutex,INFINITE);
  }

  if (STATE_KILL!=p->state&&0<q->count) {
    n=list_read(&q->active);
    *pkt=n->pkt;
#if defined (QUEUE_FREE) // {
    list_write(&q->free,n);
#else // } {
    list_write(q->free,n);
#endif // }
	--q->count;

    state=0;
  }

//fprintf(stderr,"%s read: %d, state: %d.\n",q->name,q->count,p->state);
  SetEvent(p->hEventWrite);
  ReleaseMutex(p->hMutex);

  return state;
}
