/*
 * samples.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

///////////////////////////////////////////////////////////////////////////////
typedef uint8_t u8_t;
typedef int16_t s16_t;
typedef int32_t s32_t;
typedef float f32_t;
typedef double f64_t;

///////////////////////////////////////////////////////////////////////////////
samples_t *samples_init(samples_t *s)
{
  s->hMutex=CreateMutex(
    NULL,   // _In_opt_  LPSECURITY_ATTRIBUTES lpMutexAttributes,
    FALSE,  // _In_      BOOL bInitialOwner,
    NULL    // _In_opt_  LPCTSTR lpName
  );

  if (NULL==s->hMutex)
    goto error;

  s->hEventWrite=CreateEvent(
    NULL,   // lpEventAttributes
    FALSE,  // bManualReset
    FALSE,  // bInitialState
    NULL    // lpName
  );

  if (NULL==s->hEventWrite)
    goto error;

  s->hEventRead=CreateEvent(
    NULL,   // lpEventAttributes
    FALSE,  // bManualReset
    FALSE,  // bInitialState
    NULL    // lpName
  );

  if (NULL==s->hEventRead)
    goto error;

  return s;
error:
  samples_cleanup(s);

  return NULL;
}

samples_t *samples_cleanup(samples_t *s)
{
  if (NULL!=s->rep)
    free(s->rep);

  if (NULL!=s->hEventRead)
    CloseHandle(s->hEventRead);

  if (NULL!=s->hEventWrite)
    CloseHandle(s->hEventWrite);

  if (NULL!=s->hMutex)
    CloseHandle(s->hMutex);

  return s;
}

void samples_kill(samples_t *s)
{
  WaitForSingleObject(s->hMutex,INFINITE);

  if (STATE_RUN==s->state) {
    s->wp=s->rep;
    s->state=STATE_KILL;
    SetEvent(s->hEventRead);
    SetEvent(s->hEventWrite);
  }

  ReleaseMutex(s->hMutex);
}

void samples_clear(samples_t *s)
{
  s->wp=s->rep;
  s->time=0.0;
  s->state=STATE_RUN;
}

samples_t *samples_realloc(samples_t *s, track_t *t, int nch, int bps)
{
  size_t size;
  char *rep;

  size=576*nch*(bps/8);

  if (0==s->size) {
    if (NULL==(s->rep=malloc(2*size)))
      goto error;

    s->size=size;
  }
  else if (s->size<size) {
    if (NULL==(rep=realloc(s->rep,2*size)))
      goto error;

    s->rep=rep;
    s->size=size;
  }

  s->bps=bps;
  s->t=t;
  s->mp=s->rep+size;
  s->q=track_get_q(t);

  return s;
error:
  return NULL;
}

///////////////////////////////////////////////////////////////////////////////
void samples_wait_lock(samples_t *s)
{
  WaitForSingleObject(s->hMutex,INFINITE);

  while (STATE_RUN==s->state&&s->wp==s->mp) {
    SignalObjectAndWait(s->hMutex,s->hEventWrite,INFINITE,FALSE);
    WaitForSingleObject(s->hMutex,INFINITE);
  }
}

void samples_notify_unlock(samples_t *s)
{
  SetEvent(s->hEventRead);
  ReleaseMutex(s->hMutex);
}

///////////////////////////////////////////////////////////////////////////////
#define samples_scale_s16(s) \
  (16==(s)->bps?(s)->q:(s)->q*(1<<8))
#define samples_scale_s32(s) \
  ((s)->q/(16==(s)->bps?1<<16:1<<8))
#define samples_scale_f32(s) \
  ((s)->q*(16==(s)->bps?1<<15:1<<23))
#define samples_scale_f64(s) \
  ((s)->q*(16==(s)->bps?1<<15:1<<23))

///////////////////////////////////////////////////////////////////////////////
#define SAMPLES_COPY_I(id) \
void samples_copy_##id##i(samples_t *s, uint8_t *rp, size_t nch, size_t n) \
{ \
  char *mwp; \
  char *wp; \
  id##_t x; \
  int32_t y; \
  double q=samples_scale_##id(s); \
  uint8_t *mrp=rp+nch*n*sizeof x; \
  size_t bytes=s->bps>>3; \
 \
  while (STATE_RUN==s->state&&rp<mrp) { \
	samples_wait_lock(s); \
    wp=(char *)s->wp; \
    mwp=s->mp; \
 \
    while (STATE_KILL!=s->state&&rp<mrp&&wp<mwp) { \
      memcpy(&x,rp,sizeof x); \
      rp+=sizeof x; \
 \
      y=round(q*x); \
      memcpy(wp,&y,bytes); \
      wp+=bytes; \
    } \
 \
    s->wp=wp; \
	samples_notify_unlock(s); \
  } \
}

SAMPLES_COPY_I(s16)
SAMPLES_COPY_I(s32)
SAMPLES_COPY_I(f32)
SAMPLES_COPY_I(f64)

///////////////////////////////////////////////////////////////////////////////
#define SAMPLES_COPY_P(id) \
void samples_copy_##id##p(samples_t *s, uint8_t **rp, size_t nch, size_t n) \
{ \
  char *mwp; \
  char *wp; \
  id##_t x; \
  int32_t y; \
  size_t i=0,ch; \
  double q=samples_scale_##id(s); \
  size_t bytes=s->bps>>3; \
 \
  while (STATE_RUN==s->state&&i<n) { \
	samples_wait_lock(s); \
    wp=(char *)s->wp; \
    mwp=s->mp; \
 \
    while (STATE_KILL!=s->state&&i<n&&wp<mwp) { \
      for (ch=0;ch<nch;++ch) { \
        memcpy(&x,rp[ch]+i*sizeof x,sizeof x); \
        y=round(q*x); \
        memcpy(wp,&y,bytes); \
        wp+=bytes; \
      } \
 \
      ++i; \
    } \
 \
    s->wp=wp; \
	samples_notify_unlock(s); \
  } \
}

SAMPLES_COPY_P(s16)
SAMPLES_COPY_P(s32)
SAMPLES_COPY_P(f32)
SAMPLES_COPY_P(f64)

///////////////////////////////////////////////////////////////////////////////
void samples_copy(samples_t *s, ar_t *ar)
{
  AVFrame *frame=ar->frame;
  int nch=av_frame_get_channels(frame);

  switch (frame->format) {
  // interleaved:
  case AV_SAMPLE_FMT_S16:
    samples_copy_s16i(s,frame->data[0],nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_S32:
    samples_copy_s32i(s,frame->data[0],nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_FLT:
    samples_copy_f32i(s,frame->data[0],nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_DBL:
    samples_copy_f64i(s,frame->data[0],nch,frame->nb_samples);
    break;
  // planar:
  case AV_SAMPLE_FMT_S16P:
    samples_copy_s16p(s,frame->data,nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_S32P:
    samples_copy_s32p(s,frame->data,nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_FLTP:
    samples_copy_f32p(s,frame->data,nch,frame->nb_samples);
    break;
  case AV_SAMPLE_FMT_DBLP:
    samples_copy_f64p(s,frame->data,nch,frame->nb_samples);
    break;
  // unsupported:
  default:
#if defined (DEBUG) // {
    fprintf(stderr,"Error: Unsupported sample format: \"%s\"\n",
        av_get_sample_fmt_name(frame->format));
#endif // }
    break;
  }

  avcodec_get_frame_defaults(frame);
}
