/*
 * sox_reader.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

///////////////////////////////////////////////////////////////////////////////
typedef uint8_t u8_t;
typedef int16_t s16_t;
typedef int32_t s32_t;
typedef float f32_t;
typedef double f64_t;

typedef struct priv {
  audio_t *a;

  union {
    // interleaved
    struct {
      uint8_t *rp;
      uint8_t *mrp;
    } i;

    // planar
    struct {
      uint8_t **rp;
      size_t nch;
      size_t n;
      size_t i;
    } p;
  };
} priv_t;

///////////////////////////////////////////////////////////////////////////////
static void priv_reset_i(priv_t *priv)
{
  AVFrame *frame=priv->a->frame;
  size_t n=frame->nb_samples;
  size_t nch=av_frame_get_channels(frame);
  size_t bytes=av_get_bytes_per_sample(frame->format);
  uint8_t *rp=frame->data[0];

  priv->i.rp=rp;
  priv->i.mrp=rp+n*nch*bytes;
}

static void priv_reset_p(priv_t *priv)
{
  AVFrame *frame=priv->a->frame;

  priv->p.rp=frame->data;
  priv->p.nch=av_frame_get_channels(frame);
  priv->p.n=frame->nb_samples;
  priv->p.i=0;
}

void priv_reset(priv_t *priv)
{
  AVFrame *frame=priv->a->frame;

  switch (frame->format) {
  // interleaved:
  case AV_SAMPLE_FMT_U8:
  case AV_SAMPLE_FMT_S16:
  case AV_SAMPLE_FMT_S32:
  case AV_SAMPLE_FMT_FLT:
  case AV_SAMPLE_FMT_DBL:
    priv_reset_i(priv);
    break;
  // planar:
  case AV_SAMPLE_FMT_U8P:
  case AV_SAMPLE_FMT_S16P:
  case AV_SAMPLE_FMT_S32P:
  case AV_SAMPLE_FMT_FLTP:
  case AV_SAMPLE_FMT_DBLP:
    priv_reset_p(priv);
    break;
  // unsupported:
  default:
    break;
  }
}

///////////////////////////////////////////////////////////////////////////////
static int priv_getopts(sox_effect_t *effp, int argc, char *argv[])
{
  priv_t *priv=effp->priv;
  int n=1;

  if (argc<2)
    return SOX_EOF;

  ZERO(priv);
  priv->a=(audio_t *)argv[n++];

  return SOX_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
#define priv_scale_s16(q) \
 (-(q)/INT16_MIN)
#define priv_scale_s32(q) \
 (-(q)/INT32_MIN)
#define priv_scale_f32(q) \
 (q)
#define priv_scale_f64(q) \
 (q)

///////////////////////////////////////////////////////////////////////////////
#define PRIV_DRAIN_I(id,ID) \
static int priv_drain_##id##i(sox_effect_t *effp, sox_sample_t *obuf, \
    size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  audio_t *a=priv->a; \
  ar_t *ar=&a->ar; \
  double q1=a->s.q; \
  double q2=priv_scale_##id(q1); \
  uint8_t *rp=priv->i.rp; \
  uint8_t *mrp=priv->i.mrp; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+*osamp; \
  id##_t x; \
  double y; \
  SOX_SAMPLE_LOCALS; \
 \
  while (wp<mwp) { \
    while (rp==mrp) { \
      ar_next(ar); \
 \
      if (!ar_valid(ar)) \
        break; \
 \
      priv_reset_i(priv); \
      rp=priv->i.rp; \
      mrp=priv->i.mrp; \
    } \
 \
    memcpy(&x,rp,sizeof x); \
    rp+=sizeof x; \
 \
    if (1.0==q1) \
      *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
    else { \
      y=q2*x; \
      *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
    } \
  } \
  \
  priv->i.rp=rp; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}

PRIV_DRAIN_I(s16,SIGNED_16BIT)
PRIV_DRAIN_I(s32,SIGNED_32BIT)
PRIV_DRAIN_I(f32,FLOAT_32BIT)
PRIV_DRAIN_I(f64,FLOAT_64BIT)

///////////////////////////////////////////////////////////////////////////////
#define PRIV_DRAIN_P(id,ID) \
static int priv_drain_##id##p(sox_effect_t *effp, sox_sample_t *obuf, \
    size_t *osamp) \
{ \
  priv_t *priv=effp->priv; \
  audio_t *a=priv->a; \
  ar_t *ar=&a->ar; \
  double q1=a->s.q; \
  double q2=priv_scale_##id(q1); \
  uint8_t **rp=priv->p.rp; \
  int nch=priv->p.nch; \
  int n=priv->p.n; \
  int i=priv->p.i; \
  sox_sample_t *wp=obuf; \
  sox_sample_t *mwp=obuf+*osamp; \
  int ch; \
  id##_t x; \
  double y; \
  SOX_SAMPLE_LOCALS; \
 \
  while (wp<mwp) { \
    while (i==n) { \
      ar_next(ar); \
 \
      if (!ar_valid(ar)) \
        break; \
 \
      priv_reset_p(priv); \
      rp=priv->p.rp; \
      nch=priv->p.nch; \
      n=priv->p.n; \
      i=priv->p.i; \
	} \
 \
	for (ch=0;ch<nch;++ch) { \
      memcpy(&x,rp[ch]+i*sizeof x,sizeof x); \
 \
      if (1.0==q1) \
        *wp++=SOX_##ID##_TO_SAMPLE(x,effp->clips); \
      else { \
        y=q2*x; \
        *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(y,effp->clips); \
      } \
	} \
 \
	++i; \
  } \
 \
  priv->p.i=i; \
 \
  return 0<(*osamp=wp-obuf)?SOX_SUCCESS:SOX_EOF; \
}

PRIV_DRAIN_P(s16,SIGNED_16BIT)
PRIV_DRAIN_P(s32,SIGNED_32BIT)
PRIV_DRAIN_P(f32,FLOAT_32BIT)
PRIV_DRAIN_P(f64,FLOAT_64BIT)

///////////////////////////////////////////////////////////////////////////////
static int priv_drain(sox_effect_t *effp, sox_sample_t *obuf, size_t *osamp)
{
  priv_t *priv=effp->priv;
  AVFrame *frame=priv->a->frame;

  switch (frame->format) {
  // interleaved.
  case AV_SAMPLE_FMT_S16:
    return priv_drain_s16i(effp,obuf,osamp);
  case AV_SAMPLE_FMT_S32:
    return priv_drain_s32i(effp,obuf,osamp);
  case AV_SAMPLE_FMT_FLT:
    return priv_drain_f32i(effp,obuf,osamp);
  case AV_SAMPLE_FMT_DBL:
    return priv_drain_f64i(effp,obuf,osamp);
  // planar.
  case AV_SAMPLE_FMT_S16P:
    return priv_drain_s16p(effp,obuf,osamp);
  case AV_SAMPLE_FMT_S32P:
    return priv_drain_s32p(effp,obuf,osamp);
  case AV_SAMPLE_FMT_FLTP:
    return priv_drain_f32p(effp,obuf,osamp);
  case AV_SAMPLE_FMT_DBLP:
    return priv_drain_f64p(effp,obuf,osamp);
  default:
    return SOX_EOF;
  }
}

///////////////////////////////////////////////////////////////////////////////
sox_effect_handler_t const *ffsox_reader() {
  static sox_effect_handler_t handler={
    // Effect name
    .name="ffsox_reader",
    // Short explanation of parameters accepted by effect
    .usage=NULL,
    // Combination of SOX_EFF_* flags
    .flags=SOX_EFF_MCHAN,
    // Called to parse command-line arguments (called once per effect).
    .getopts=priv_getopts,
    // Called to initialize effect (called once per flow).
    .start=NULL,
    // Called to process samples.
    .flow=NULL,
    // Called to finish getting output after input is complete.
    .drain=priv_drain,
    // Called to shut down effect (called once per flow).
    .stop=NULL,
    // Called to shut down effect (called once per effect).
    .kill=NULL,
    // Size of private data SoX should pre-allocate for effect
    .priv_size=sizeof (priv_t)
  };

  return &handler;
}
